"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Try = void 0;
const NoSuchElementException_1 = require("../exceptions/NoSuchElementException");
const TryFunctions = {
    OF: "OF",
    MAP: 'MAP',
    ANDTHEN: 'ANDTHEN',
    FLATMAP: 'FLATMAP',
    FILTER: 'FILTER',
    FILTERNOT: 'FILTERNOT',
    PEEK: 'PEEK',
    RECOVER: 'RECOVER',
    RECOVERWITH: 'RECOVERWITH',
    ONSUCCESS: 'ONSUCCESS',
    ONFAILURE: 'ONFAILURE',
};
class Try {
    constructor(initExecution) {
        this.executionStack = [];
        if (initExecution)
            this.executionStack.push(initExecution);
    }
    //Static methods to create a Try object
    static success(value) {
        return new Try().setValue(value);
    }
    static failure(error) {
        return new Try().setError(error);
    }
    static of(fn) {
        return new Try({
            name: TryFunctions.OF,
            functionData: { func: fn },
            returning: true
        });
    }
    //Private methods to modify the internal state
    setValue(value) {
        this.value = value;
        return this;
    }
    setError(err) {
        this.errorStack = err;
        return this;
    }
    runElement(executionElement_1) {
        return __awaiter(this, arguments, void 0, function* (executionElement, isFirst = false) {
            try {
                if (isFirst && executionElement.returning) {
                    this.value = yield executionElement.functionData.func();
                    return;
                }
                if (executionElement.returning) {
                    this.value = yield executionElement.functionData.func(this.value);
                    return;
                }
                yield executionElement.functionData.func(this.value);
            }
            catch (e) {
                this.errorStack = e;
            }
        });
    }
    runExecutionStack() {
        return __awaiter(this, void 0, void 0, function* () {
            for (let i = 0; i < this.executionStack.length; i++) {
                const executionElement = this.executionStack[i];
                try {
                    switch (executionElement.name) {
                        case TryFunctions.MAP: {
                            if (this.isSuccess())
                                yield this.runElement(executionElement, executionElement.name === TryFunctions.OF);
                            break;
                        }
                        case TryFunctions.FLATMAP: {
                            if (this.isSuccess()) {
                                const tryObject = yield executionElement.functionData.func(this.value);
                                this.value = yield tryObject.get();
                            }
                            break;
                        }
                        case TryFunctions.ANDTHEN: {
                            if (this.isSuccess())
                                yield this.runElement(executionElement, executionElement.name === TryFunctions.OF);
                            break;
                        }
                        case TryFunctions.FILTER: {
                            if (this.isSuccess()) {
                                if (!(yield executionElement.functionData.func(this.value))) {
                                    yield executionElement.functionData.fallbackFunction(this.value);
                                }
                            }
                            break;
                        }
                        case TryFunctions.FILTERNOT: {
                            if (this.isSuccess()) {
                                if (yield executionElement.functionData.func(this.value)) {
                                    yield executionElement.functionData.fallbackFunction(this.value);
                                }
                            }
                            break;
                        }
                        case TryFunctions.PEEK: {
                            if (this.isSuccess())
                                yield executionElement.functionData.func(this.value);
                            break;
                        }
                        case TryFunctions.RECOVER: {
                            if (this.isFailure()) {
                                this.errorStack = undefined;
                                this.value = yield executionElement.functionData.func(this.errorStack);
                            }
                            break;
                        }
                        case TryFunctions.RECOVERWITH: {
                            if (this.isFailure()) {
                                this.errorStack = undefined;
                                const tryObject = yield executionElement.functionData.func(this.errorStack);
                                this.value = yield tryObject.get();
                            }
                            break;
                        }
                        default: {
                            //This will typically run one of the static methods
                            yield this.runElement(executionElement, executionElement.name === TryFunctions.OF);
                        }
                    }
                }
                catch (ex) {
                    // @ts-ignore
                    this.errorStack = ex;
                }
            }
        });
    }
    //----- Public interface -----
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.runExecutionStack();
            return this;
        });
    }
    get() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.runExecutionStack();
            if (this.isFailure())
                throw this.errorStack;
            return this.value;
        });
    }
    getOrElse(defaultValue) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.runExecutionStack();
            return this.isFailure() ? defaultValue : this.value;
        });
    }
    getOrElseGet(fn) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.runExecutionStack();
            return this.isFailure() ? yield fn(this.errorStack) : this.value;
        });
    }
    getOrElseThrow(fn) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.runExecutionStack();
            if (this.isFailure())
                throw fn(this.errorStack);
            return this.value;
        });
    }
    map(fn) {
        this.executionStack.push({
            name: TryFunctions.MAP,
            functionData: { func: fn },
            returning: true
        });
        return this;
    }
    flatMap(fn) {
        this.executionStack.push({
            name: TryFunctions.FLATMAP,
            functionData: { func: fn },
            returning: true
        });
        return this;
    }
    filter(predicateFunc, throwbackFunction) {
        this.executionStack.push({
            name: TryFunctions.FILTER,
            functionData: { func: predicateFunc, fallbackFunction: throwbackFunction !== null && throwbackFunction !== void 0 ? throwbackFunction : ((value) => { throw new NoSuchElementException_1.NoSuchElementException(`Predicate does not hold for ${value}`); }) },
            returning: false
        });
        return this;
    }
    filterNot(predicateFunc, throwbackFunction) {
        this.executionStack.push({
            name: TryFunctions.FILTERNOT,
            functionData: { func: predicateFunc, fallbackFunction: throwbackFunction !== null && throwbackFunction !== void 0 ? throwbackFunction : ((value) => { throw new NoSuchElementException_1.NoSuchElementException(`Predicate does not hold for ${value}`); }) },
            returning: false
        });
        return this;
    }
    peek(fn) {
        this.executionStack.push({
            name: TryFunctions.PEEK,
            functionData: { func: fn },
            returning: false
        });
        return this;
    }
    andThen(fn) {
        this.executionStack.push({
            name: TryFunctions.ANDTHEN,
            functionData: { func: fn },
            returning: false
        });
        return this;
    }
    recover(fn) {
        this.executionStack.push({
            name: TryFunctions.RECOVER,
            functionData: { func: fn },
            returning: true
        });
        return this;
    }
    recoverWith(fn) {
        this.executionStack.push({
            name: TryFunctions.RECOVERWITH,
            functionData: { func: fn },
            returning: true
        });
        return this;
    }
    isSuccess() {
        return this.errorStack === undefined;
    }
    isFailure() {
        return this.errorStack !== undefined;
    }
    onSuccess(fn) {
        this.executionStack.push({
            name: TryFunctions.ONSUCCESS,
            functionData: { func: fn },
            returning: false
        });
        return this;
    }
    onFailure(fn) {
        this.executionStack.push({
            name: TryFunctions.ONFAILURE,
            functionData: { func: fn },
            returning: false
        });
        return this;
    }
    getCause() {
        return this.errorStack;
    }
}
exports.Try = Try;
